package com.virnect.tracksample

import android.Manifest
import android.app.AlertDialog
import android.content.DialogInterface
import android.content.Intent
import android.content.pm.PackageManager
import android.net.Uri
import android.os.Build
import android.os.Bundle
import android.os.Handler
import android.text.InputType
import android.util.Log
import android.view.View
import android.view.ViewGroup
import android.widget.CheckBox
import android.widget.EditText
import android.widget.LinearLayout
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.appcompat.widget.AppCompatCheckedTextView
import androidx.core.app.ActivityCompat
import androidx.core.content.ContextCompat
import androidx.core.view.get
import com.virnect.FileAccessHelper
import kotlinx.android.synthetic.main.activity_main.*
import java.io.IOException
import java.text.SimpleDateFormat
import java.util.*
import java.util.stream.IntStream.range


class MainActivity : AppCompatActivity() {

    //region Initialization
    private val TAG = "TRACK_MENU" ///< Log tag

    private var trackConfig = TrackConfig.instance() ///< Shared configuration singleton

    private val PERMISSIONSTARTOK = 100 ///< Permission request start code
    private val PERMISSIONPLAYOK = 101  ///< Permission request play code

    private var doubleBackToExitPressedOnce = false ///< Double press to exit flag

    private val supportedDevices = listOf(
        "SM-G991",
        "SM-G988",
        "SM-G977",
        "SM-N976"
    ) ///< Officially supported device by TRACK

    /**
     * Reads the intent extra and sets the APP configuration accordingly
     */
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        setContentView(R.layout.activity_main)

        // IMG + QR
        buttonStartImageTracking.setOnClickListener { setupAndStartTrackFramework(PlayMode.Image) }
        buttonStartQRCodeTracking.setOnClickListener { setupAndStartTrackFramework(PlayMode.QR) }
        buttonStartHybrid.setOnClickListener { setupAndStartTrackFramework(PlayMode.Hybrid) }

        // CAD
        buttonStartCADTargetTracking.setOnClickListener {
            val targets = FileAccessHelper.listTargetNames(this, trackConfig.targetDataPathCAD)

            // setup selection dialog
            val builder = AlertDialog.Builder(this)
            builder.setTitle("Select Target")
            builder.setItems(targets) { _, index ->
                trackConfig.targetCAD = arrayOf(targets[index])
                setupAndStartTrackFramework(PlayMode.CAD)
            }
            val dialog = builder.create()
            dialog.show()
        }

        // MAP
        buttonStartMapRecording.setOnClickListener {

            val format =
                SimpleDateFormat("ddMMyyyy_hhmmss") // "_" is the only special character allowed for target names
            var proposedName = "MAP_" + format.format(Date())

            // Show dialog to modify name
            val builder = AlertDialog.Builder(this)
            builder.setTitle("New MAP target name")
            val input = EditText(this)
            input.inputType = InputType.TYPE_CLASS_TEXT
            input.setText(proposedName)

            builder.setView(input)
            builder.setPositiveButton("OK") { _, _ ->
                proposedName = input.text.toString()
                //Sanitize input
                if (proposedName.isEmpty()) {
                    Toast.makeText(this, "Target name can not be empty", Toast.LENGTH_SHORT).show()
                    return@setPositiveButton
                }
                val regex = "^[a-zA-Z0-9_]+$".toRegex()
                if (!regex.matches(proposedName)) {
                    Toast.makeText(
                        this,
                        "Target name can only consist of 'a-z' 'A-Z' '0-9' '_'",
                        Toast.LENGTH_SHORT
                    ).show()
                    return@setPositiveButton
                }
                val existingTargets =
                    FileAccessHelper.listTargetNames(this, trackConfig.targetDataPathMap)
                if (existingTargets.contains(proposedName)) {
                    Toast.makeText(this, "Target name already exists", Toast.LENGTH_SHORT).show()
                    return@setPositiveButton
                }

                //Name is valid
                trackConfig.mapRecordingTargetName = proposedName
                setupAndStartTrackFramework(PlayMode.MapRecording)
            }
            builder.setNegativeButton("Cancel") { dialog, _ -> dialog.cancel() }
            builder.show()
        }

        buttonStartMapTargetTracking.setOnClickListener {

            val targets = FileAccessHelper.listTargetNames(this, trackConfig.targetDataPathMap)

            if (targets == null || targets.isEmpty()) {
                Toast.makeText(this, "No MAP targets recorded yet.", Toast.LENGTH_SHORT).show()
                return@setOnClickListener
            }

            // Setup selection dialog
            val builder = AlertDialog.Builder(this)
            builder.setTitle("Select Targets")

            val checkedItems = BooleanArray(targets.size)
            builder.setMultiChoiceItems(targets, checkedItems) { dialog, index, isChecked ->
                // The user checked or unchecked a box
                if (checkedItems.count { it } >= trackConfig.maxAmountOfMapTargets) {
                    // If the user selects more than 3 items, disable all checkboxes that are not checked
                    for (i in range(0, targets.size)) {

                        val item =
                            ((dialog as AlertDialog).listView.get(i) as AppCompatCheckedTextView)
                        item.isEnabled = checkedItems[i]

                        // Since the onClicklistener of disabled checkboxes is still active
                        // we need to suppress the explicit checking of an entry
                        if (checkedItems.count { it } > trackConfig.maxAmountOfMapTargets && index == i) {
                            // Update data
                            checkedItems[index] = false
                            dialog.listView.setItemChecked(index, false)

                            item.isEnabled = false // update UI

                            Toast.makeText(
                                this,
                                "You can not select more than ${trackConfig.maxAmountOfMapTargets} map targets.",
                                Toast.LENGTH_LONG
                            ).show()
                        }
                    }
                } else {
                    // Enable all check boxes
                    for (i in range(0, targets.size)) {
                        val item =
                            ((dialog as AlertDialog).listView.get(i) as AppCompatCheckedTextView)
                        item.isEnabled = true
                    }
                }
            }

            // Add OK and Cancel buttons
            builder.setPositiveButton("OK") { _, _ ->
                // The user clicked OK - generate target list
                val map = targets.filterIndexed { index, _ -> checkedItems[index] }.toTypedArray()

                if (map.isEmpty()) {
                    Toast.makeText(
                        this,
                        "At least one target should be selected",
                        Toast.LENGTH_SHORT
                    ).show()

                    return@setPositiveButton
                }

                trackConfig.targetMap = map
                setupAndStartTrackFramework(PlayMode.MapReplay)
            }
            builder.setNegativeButton("Cancel", null)

            // Create and show the alert dialog
            val dialog = builder.create()
            dialog.show()
        }

        // About Button
        buttonInstructions.setOnClickListener {
            val builder: AlertDialog.Builder = AlertDialog.Builder(this)
            builder.setTitle("Instructions")
            builder.setMessage(R.string.instructionText)
            builder.setPositiveButton("OK", null)
            builder.setNegativeButton("View Documentation") { _: DialogInterface, _: Int ->
                val browserIntent = Intent(
                    Intent.ACTION_VIEW,
                    Uri.parse(getString(R.string.linkTrackDocumentation))
                )
                startActivity(browserIntent)
            }
            val dialog: AlertDialog = builder.create()
            dialog.show()
        }

        // If model is not supported by TRACK, show warning dialog
        val sharedPref = getPreferences(MODE_PRIVATE)
        if (sharedPref.getBoolean(getString(R.string.unsupportedDeviceDialog), true)) {
            if (supportedDevices.filter { Build.MODEL.contains(it) }.isEmpty()) {
                val builder = AlertDialog.Builder(this)
                builder.setTitle("Warning")

                val layout = LinearLayout(this)
                val params = LinearLayout.LayoutParams(
                    ViewGroup.LayoutParams.WRAP_CONTENT,
                    ViewGroup.LayoutParams.WRAP_CONTENT
                )
                params.setMargins(45, 45, 0, 0)

                val checkbox = CheckBox(this)
                checkbox.text = "Do not show again"

                layout.addView(checkbox, params)
                builder.setView(layout)
                builder.setMessage("This device is not officially supported by TRACK. You can still use TRACK but tracking quality will not be guaranteed.")
                builder.setPositiveButton("OK") { _, _ ->
                    if (checkbox.isChecked)
                        with(sharedPref.edit()) {
                            putBoolean(getString(R.string.unsupportedDeviceDialog), false)
                            commit()
                        }
                }
                builder.setNegativeButton("QUIT") { _, _ -> finish() }
                builder.show()
            }
        }
    }

    //endregion

    //region Android lifecycle integration

    /**
     * Prepare target data when the app starts
     */
    override fun onStart() {
        super.onStart()
        permissionCheck(PERMISSIONSTARTOK)

        // Execute file access in separate IO thread
        Thread {
            // Prepare target data files
            if (!prepareTargetData()) {
                Toast.makeText(
                    this,
                    "TRACK : ERROR Could not prepare target data",
                    Toast.LENGTH_LONG
                ).show()
            }
        }.start()
    }

    /**
     * Start processing when the app is active
     */
    override fun onResume() {
        super.onResume()

        // Always hide loading screen on startup
        loadingScreen.visibility = View.GONE

        // In case of fileSequence input, an external processing thread
        // is automatically started after initialization
        if (trackConfig.useFileSequence)
            setupAndStartTrackFramework(trackConfig.playMode)
    }
    //endregion

    //region Fullscreen and Backpress

    /**
     * Handle back press
     */
    override fun onBackPressed() {
        // Handle double back press
        if (doubleBackToExitPressedOnce) {
            finish()
            return
        }
        doubleBackToExitPressedOnce = true
        Toast.makeText(this, "Please click BACK again to exit", Toast.LENGTH_SHORT).show()
        Handler().postDelayed({ doubleBackToExitPressedOnce = false }, 2000)
    }

    /**
     * Call fullscreen
     */
    override fun onWindowFocusChanged(hasFocus: Boolean) {
        super.onWindowFocusChanged(hasFocus)
        if (hasFocus) hideSystemUI()
    }

    /**
     * Apply fullscreen
     */
    @Suppress("DEPRECATION")
    private fun hideSystemUI() {
        window.decorView.systemUiVisibility = (
                View.SYSTEM_UI_FLAG_LAYOUT_STABLE
                        or View.SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION
                        or View.SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN
                        or View.SYSTEM_UI_FLAG_HIDE_NAVIGATION
                        or View.SYSTEM_UI_FLAG_FULLSCREEN
                        or View.SYSTEM_UI_FLAG_IMMERSIVE
                )
    }

    //endregion

    //region Permission Helper

    /**
     * Helper to perform permission checkup
     */
    private fun permissionCheck(requestCode: Int): Boolean {
        if (ContextCompat.checkSelfPermission(
                this,
                Manifest.permission.CAMERA
            ) != PackageManager.PERMISSION_GRANTED ||
            ContextCompat.checkSelfPermission(
                this,
                Manifest.permission.READ_EXTERNAL_STORAGE
            ) != PackageManager.PERMISSION_GRANTED ||
            ContextCompat.checkSelfPermission(
                this,
                Manifest.permission.WRITE_EXTERNAL_STORAGE
            ) != PackageManager.PERMISSION_GRANTED
        ) {
            ActivityCompat.requestPermissions(
                this,
                arrayOf(
                    Manifest.permission.CAMERA,
                    Manifest.permission.READ_EXTERNAL_STORAGE,
                    Manifest.permission.WRITE_EXTERNAL_STORAGE
                ),
                requestCode
            )
            return false
        }
        return true
    }

    /**
     * Permission result callback
     */
    override fun onRequestPermissionsResult(
        requestCode: Int,
        permissions: Array<out String>,
        grantResults: IntArray
    ) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults)
        if (requestCode == PERMISSIONPLAYOK &&
            grantResults[0] == PackageManager.PERMISSION_GRANTED &&
            grantResults[1] == PackageManager.PERMISSION_GRANTED &&
            grantResults[2] == PackageManager.PERMISSION_GRANTED
        ) {
            setupAndStartTrackFramework(trackConfig.playMode)
        } else {
            Toast.makeText(
                this,
                "All permissions need to be granted for using VIRNECT TRACK Framework",
                Toast.LENGTH_LONG
            ).show()
        }
    }

    //endregion

    //region Track framework initialization

    /**
     * Initialize framework if permission is granted
     */
    private fun setupAndStartTrackFramework(mode: PlayMode) {
        // Setup config
        trackConfig.setup(mode)

        // Check permission, callback will call this method again
        if (!permissionCheck(PERMISSIONPLAYOK)) {
            Log.e(TAG, "Can not initialize framework. No camera or file permission")
            return
        }

        if (trackConfig.targets == null) {
            Toast.makeText(this, "ERROR No target set", Toast.LENGTH_LONG).show()
            return
        }

        // Show loading screen
        loadingScreen.visibility = View.VISIBLE

        // Start tracking activity
        startActivity(Intent(this, TrackActivity::class.java))
    }

    /**
     * Prepares the target data
     * In case of normal usage, extract the embedded files with the FileAccessHelper
     * In case of fileSequence, set the given path
     */
    private fun prepareTargetData(): Boolean {
        try {
            // Copy target files from APP assets to cache directory
            FileAccessHelper.copyTargetData(this, trackConfig.targetDataPathStandard)

            // Prepare map target folder
            val mapTargetPath = FileAccessHelper.getLocalPath(this, trackConfig.targetDataPathMap)
            if (!mapTargetPath.exists())
                if (!mapTargetPath.mkdirs())
                    throw IOException("Could not create folder")

            return true
        } catch (e: IOException) {
            e.printStackTrace()
            Toast.makeText(
                applicationContext,
                "TRACK : ERROR could not prepare target data files",
                Toast.LENGTH_LONG
            ).show()
            return false
        }
    }
    //endregion
}
