package com.virnect.tracksample.Renderer;

import android.opengl.GLES32;
import android.opengl.Matrix;
import android.util.Log;

import com.virnect.Common;
import com.virnect.Track;

import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.ArrayList;
import java.util.List;

import javax.microedition.khronos.opengles.GL10;

/**
 * Renderable class
 * This class is for rendering tracking targets
 */
public class Renderable {

    /**
     * RenderType has 2 types
     * Only plane can render texture on it
     */
    public enum RenderType {
        Point, ///< Map points
        Plane, ///< Plane which is attached to target,
        Model, ///< CAD Model
        Axis   ///< Target's axis
    }

    /**
     * List of all Renderable instances
     */
    private static List<Renderable> allRenderables = new ArrayList<>();

    /**
     * Rotation to apply in vertex shader
     */
    private static float screenRotation = 0;

    /**
     * Initialized flag
     */
    private boolean isInitialized = false;

    /**
     * Static method to apply screen rotation settings to all renderables
     *
     * @param rotation rotation to set
     */
    public static void applyScreenRotation(float rotation) {
        screenRotation = rotation;
        for (Renderable r : allRenderables)
            r.applyRotation();
    }

    /**
     * Initialize Renderable with RenderType
     *
     * @param type       If place is set to initialize function, Renderable will render place shape
     * @param hasTexture Texture rendering only if type is Plane
     */
    public boolean initialize(final RenderType type, final boolean hasTexture) {
        if (type == RenderType.Axis && hasTexture)
            Log.w("Renderable", "Axis type does not support texture");

        if (isInitialized)
            return false;

        mRenderableShader = new ShaderManager(vertexShaderSource, fragmentShaderSource);

        IntBuffer vaoBuffer = IntBuffer.allocate(1);
        IntBuffer vboBuffer = IntBuffer.allocate(1);
        IntBuffer eboBuffer = IntBuffer.allocate(1);

        GLES32.glGenVertexArrays(1, vaoBuffer);
        mVAO = vaoBuffer.get(0);
        GLES32.glBindVertexArray(mVAO);

        GLES32.glGenBuffers(1, vboBuffer);
        GLES32.glGenBuffers(1, eboBuffer);

        mVBO = vboBuffer.get(0);
        mEBO = eboBuffer.get(0);

        if (type == RenderType.Axis) {
            float[] vertices = {
                    0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                    0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                    0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.0f,
                    1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                    0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 0.0f, 1.0f, 0.0f, 0.0f,
                    0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.0f
            };
            int[] indices = new int[]{0, 3, 1, 4, 2, 5};

            mAmountOfIndices = indices.length;

            GLES32.glBindBuffer(GLES32.GL_ARRAY_BUFFER, mVBO);
            GLES32.glBufferData(GLES32.GL_ARRAY_BUFFER, vertices.length * Float.BYTES, FloatBuffer.wrap(vertices), GLES32.GL_STATIC_DRAW);

            GLES32.glBindBuffer(GLES32.GL_ELEMENT_ARRAY_BUFFER, mEBO);
            GLES32.glBufferData(GLES32.GL_ELEMENT_ARRAY_BUFFER, indices.length * Integer.BYTES, IntBuffer.wrap(indices), GLES32.GL_STATIC_DRAW);

            //Set line width
            GLES32.glEnable(GL10.GL_LINE_SMOOTH);
            GLES32.glLineWidth(17.5f);
        } else if (type == RenderType.Plane){
            if (hasTexture) {
                // Default is Plane
                float[] vertices = {
                        1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 0.0f,
                        1.0f, -1.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f,
                        -1.0f, -1.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f, 0.0f, 1.0f,
                        -1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f, 0.0f, 0.0f};

                GLES32.glBindBuffer(GLES32.GL_ARRAY_BUFFER, mVBO);
                GLES32.glBufferData(GLES32.GL_ARRAY_BUFFER, vertices.length * Float.BYTES, FloatBuffer.wrap(vertices), GLES32.GL_STATIC_DRAW);

                IntBuffer textureID = IntBuffer.allocate(1);
                GLES32.glGenTextures(1, textureID);

                mTextureID = textureID.get(0);
                GLES32.glTexImage2D(GLES32.GL_TEXTURE_2D, 0, GLES32.GL_RGB, Track.getPreviewSize()[0], Track.getPreviewSize()[1], 0, GLES32.GL_RGBA, GLES32.GL_UNSIGNED_BYTE, null);

                GLES32.glTexParameteri(GLES32.GL_TEXTURE_2D, GLES32.GL_TEXTURE_WRAP_S, GLES32.GL_REPEAT);
                GLES32.glTexParameteri(GLES32.GL_TEXTURE_2D, GLES32.GL_TEXTURE_WRAP_T, GLES32.GL_REPEAT);
                GLES32.glTexParameteri(GLES32.GL_TEXTURE_2D, GLES32.GL_TEXTURE_MIN_FILTER, GLES32.GL_LINEAR);
                GLES32.glTexParameteri(GLES32.GL_TEXTURE_2D, GLES32.GL_TEXTURE_MAG_FILTER, GLES32.GL_LINEAR);

            } else {
                // Default is Plane
                float[] vertices = {
                        1.0f, 1.0f, 0.0f, 0.78f, 0.08f, 0.52f, 0.5f, 0.0f, 0.0f,
                        1.0f, -1.0f, 0.0f, 0.78f, 0.08f, 0.52f, 0.5f, 0.0f, 0.0f,
                        -1.0f, -1.0f, 0.0f, 0.78f, 0.08f, 0.52f, 0.5f, 0.0f, 0.0f,
                        -1.0f, 1.0f, 0.0f, 0.78f, 0.08f, 0.52f, 0.5f, 0.0f, 0.0f};

                GLES32.glBindBuffer(GLES32.GL_ARRAY_BUFFER, mVBO);
                GLES32.glBufferData(GLES32.GL_ARRAY_BUFFER, vertices.length * Float.BYTES, FloatBuffer.wrap(vertices), GLES32.GL_STATIC_DRAW);
            }

            int[] indices = new int[]{
                    0, 1, 3, // first triangle
                    1, 2, 3  // second triangle
            };

            GLES32.glBindBuffer(GLES32.GL_ELEMENT_ARRAY_BUFFER, mEBO);
            GLES32.glBufferData(GLES32.GL_ELEMENT_ARRAY_BUFFER, indices.length * Integer.BYTES, IntBuffer.wrap(indices), GLES32.GL_STATIC_DRAW);
            mAmountOfIndices = indices.length;
        }

        int positionAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aPos");
        // position attribute
        GLES32.glEnableVertexAttribArray(positionAttribute);
        GLES32.glVertexAttribPointer(positionAttribute, 3, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 0);

        // color attribute
        int colorAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aColor");
        GLES32.glEnableVertexAttribArray(colorAttribute);
        GLES32.glVertexAttribPointer(colorAttribute, 4, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 3 * Float.BYTES);

        // texture attribute
        int textureAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aTexCoord");
        GLES32.glEnableVertexAttribArray(textureAttribute);
        GLES32.glVertexAttribPointer(textureAttribute, 2, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 7 * Float.BYTES);

        GLES32.glBindVertexArray(0);

        mType = type;
        isInitialized = true;

        allRenderables.add(this);
        applyRotation();

        return true;
    }

    public void setIndices(int[] indices){
        GLES32.glBindVertexArray(mVAO);

        mAmountOfIndices = indices.length;

        GLES32.glBindBuffer(GLES32.GL_ELEMENT_ARRAY_BUFFER, mEBO);
        GLES32.glBufferData(GLES32.GL_ELEMENT_ARRAY_BUFFER, indices.length * Integer.BYTES, IntBuffer.wrap(indices), GLES32.GL_STATIC_DRAW);

        GLES32.glBindVertexArray(0);
    }

    /**
     * Set Renderable vertices
     * @param vertices vertices for drawing
     */
    public void setVertices(Common.Vertex[] vertices,boolean hasColor){

        GLES32.glBindVertexArray(mVAO);

        FloatBuffer drawableVertices = FloatBuffer.allocate(vertices.length * 9);

        for (Common.Vertex vertex: vertices) {

            drawableVertices.put((float)vertex.mPos.x);
            drawableVertices.put((float)vertex.mPos.y);
            drawableVertices.put((float)vertex.mPos.z);

            if(hasColor){
                drawableVertices.put((float)vertex.mRGBA.x);
                drawableVertices.put((float)vertex.mRGBA.y);
                drawableVertices.put((float)vertex.mRGBA.z);
                drawableVertices.put((float)vertex.mRGBA.w);
            }else{
                drawableVertices.put(0.78f);
                drawableVertices.put(0.08f);
                drawableVertices.put(0.52f);
                drawableVertices.put(0.5f);
            }

            drawableVertices.put(0.0f);
            drawableVertices.put(0.0f);
        }

        drawableVertices.rewind();

        GLES32.glBindBuffer(GLES32.GL_ARRAY_BUFFER, mVBO);
        GLES32.glBufferData(GLES32.GL_ARRAY_BUFFER, drawableVertices.capacity() * Float.BYTES,drawableVertices, GLES32.GL_STATIC_DRAW);

        int positionAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aPos");
        // position attribute
        GLES32.glEnableVertexAttribArray(positionAttribute);
        GLES32.glVertexAttribPointer(positionAttribute, 3, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 0);

        // color attribute
        int colorAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aColor");
        GLES32.glEnableVertexAttribArray(colorAttribute);
        GLES32.glVertexAttribPointer(colorAttribute, 4, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 3 * Float.BYTES);

        // texture attribute
        int textureAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aTexCoord");
        GLES32.glEnableVertexAttribArray(textureAttribute);
        GLES32.glVertexAttribPointer(textureAttribute, 2, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 7 * Float.BYTES);

        GLES32.glBindVertexArray(0);
    }

    /**
     * Set Renderable vertices
     * @param vertices vertices for drawing
     */
    public void setVertices(Common.Vec3[] vertices){

        GLES32.glBindVertexArray(mVAO);

        FloatBuffer drawableVertices = FloatBuffer.allocate(vertices.length * 9);

        for (Common.Vec3 vertex: vertices) {

            drawableVertices.put((float)vertex.x);
            drawableVertices.put((float)vertex.y);
            drawableVertices.put((float)vertex.z);

            drawableVertices.put(1.0f);
            drawableVertices.put(0.0f);
            drawableVertices.put(0.0f);
            drawableVertices.put(1.0f);

            drawableVertices.put(0.0f);
            drawableVertices.put(0.0f);
        }

        drawableVertices.rewind();

        mAmountOfIndices = vertices.length;

        GLES32.glBindBuffer(GLES32.GL_ARRAY_BUFFER, mVBO);
        GLES32.glBufferData(GLES32.GL_ARRAY_BUFFER, drawableVertices.capacity() * Float.BYTES,drawableVertices, GLES32.GL_STATIC_DRAW);

        int positionAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aPos");
        // position attribute
        GLES32.glEnableVertexAttribArray(positionAttribute);
        GLES32.glVertexAttribPointer(positionAttribute, 3, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 0);

        // color attribute
        int colorAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aColor");
        GLES32.glEnableVertexAttribArray(colorAttribute);
        GLES32.glVertexAttribPointer(colorAttribute, 4, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 3 * Float.BYTES);

        // texture attribute
        int textureAttribute = GLES32.glGetAttribLocation(mRenderableShader.ID, "aTexCoord");
        GLES32.glEnableVertexAttribArray(textureAttribute);
        GLES32.glVertexAttribPointer(textureAttribute, 2, GLES32.GL_FLOAT, false, 9 * Float.BYTES, 7 * Float.BYTES);

        GLES32.glBindVertexArray(0);
    }

    /**
     * Set Renderable scale. You can use target information of VIRNECT Track
     *
     * @param x X scale
     * @param y Y scale
     * @param z Z scale
     */
    public void setScale(float x, float y, float z) {
        float[] scaleMatrix = new float[16];

        scaleMatrix[0] = x;
        scaleMatrix[5] = y;
        scaleMatrix[10] = z;
        scaleMatrix[15] = 1.0f;

        mRenderableShader.use();
        int scaleLoc = GLES32.glGetUniformLocation(mRenderableShader.ID, "scale");
        GLES32.glUniformMatrix4fv(scaleLoc, 1, false, FloatBuffer.wrap(scaleMatrix));
    }

    public void setPosition(float x, float y, float z)
    {
        float[] posMatrix = new float[16];

        posMatrix[12] = x;
        posMatrix[13] = y;
        posMatrix[14] = z;

        posMatrix[0] = posMatrix[5] = posMatrix[10] = posMatrix[15] = 1.0f;

        mRenderableShader.use();
        int posLoc = GLES32.glGetUniformLocation(mRenderableShader.ID, "locPos");
        GLES32.glUniformMatrix4fv(posLoc, 1, false, FloatBuffer.wrap(posMatrix));
    }

    /**
     * Set Projection matrix using camera intrinsic parameters
     *
     * @param cameraCalibration Camera intrinsic parameter
     */
    public void setPerspective(final Common.CameraCalibration cameraCalibration) {
        float[] projectionMatrix = new float[16];
        float farPlane = 10000.0f;
        float nearPlane = 0.0001f;

        projectionMatrix[0] = 2.0f * (float) cameraCalibration.mFocalLength[0] / cameraCalibration.mResolution[0];
        projectionMatrix[5] = 2.0f * (float) cameraCalibration.mFocalLength[1] / cameraCalibration.mResolution[1];
        projectionMatrix[8] = 1.0f - 2.0f * (float) cameraCalibration.mPrincipalPoint[0] / cameraCalibration.mResolution[0];
        projectionMatrix[9] = 2.0f * (float) cameraCalibration.mPrincipalPoint[1] / cameraCalibration.mResolution[1] - 1.0f;
        projectionMatrix[10] = (farPlane + nearPlane) / (nearPlane - farPlane);
        projectionMatrix[11] = -1.0f;
        projectionMatrix[14] = 2.0f * farPlane * nearPlane / (nearPlane - farPlane);

        mRenderableShader.use();

        int perspectiveLoc = GLES32.glGetUniformLocation(mRenderableShader.ID, "perspective");
        GLES32.glUniformMatrix4fv(perspectiveLoc, 1, false, FloatBuffer.wrap(projectionMatrix));
    }

    /**
     * Set Orthogonal for background
     */
    public void setOrthogonal() {
        float[] orthogonal = new float[16];
        Matrix.frustumM(orthogonal, 0, -1, 1, -1, 1, 0.1f, 10f);

        mRenderableShader.use();

        int perspectiveLoc = GLES32.glGetUniformLocation(mRenderableShader.ID, "perspective");
        GLES32.glUniformMatrix4fv(perspectiveLoc, 1, false, FloatBuffer.wrap(orthogonal));
    }

    /**
     * Set rotation matrix
     */
    public void applyRotation() {
        float[] rotationMatrix = new float[16];
        Matrix.setIdentityM(rotationMatrix, 0);
        Matrix.rotateM(rotationMatrix, 0, screenRotation, 0.0f, 0.0f, 1.0f);
        mRenderableShader.use();
        GLES32.glUniformMatrix4fv(GLES32.glGetUniformLocation(mRenderableShader.ID, "screenRotation"), 1, false, rotationMatrix, 0);
    }

    /**
     * Render background object with texture
     *
     * @param size  Texture size
     * @param frame Preview frame data
     */
    public void render(int[] size, ByteBuffer frame) {
        if (mTextureID == 0 || mType == RenderType.Axis)
            return;

        GLES32.glActiveTexture(GLES32.GL_TEXTURE0);
        GLES32.glBindTexture(GLES32.GL_TEXTURE_2D, mTextureID);

        GLES32.glTexImage2D(GLES32.GL_TEXTURE_2D, 0, GLES32.GL_RGBA, size[0], size[1], 0, GLES32.GL_RGBA, GLES32.GL_UNSIGNED_BYTE, frame);
        GLES32.glGenerateMipmap(GLES32.GL_TEXTURE_2D);

        float[] transform = new float[16];
        Matrix.setIdentityM(transform, 0);
        Matrix.translateM(transform, 0, 0.0f, 0.0f, -0.10f);
        mRenderableShader.use();

        GLES32.glBindVertexArray(mVAO);
        int transformLoc = GLES32.glGetUniformLocation(mRenderableShader.ID, "transform");

        GLES32.glUniformMatrix4fv(transformLoc, 1, false, FloatBuffer.wrap(transform));

        GLES32.glDrawElements(GLES32.GL_TRIANGLES, 6, GLES32.GL_UNSIGNED_INT, 0);

        GLES32.glBindTexture(GLES32.GL_TEXTURE_2D, 0);
        GLES32.glBindVertexArray(0);
    }

    /**
     * Draw Renderable object
     *
     * @param result VIRNECT ObjectTrackResult instance which has target's transformation information
     */
    public void render(final Common.TrackerResult result) {
        float[] transform = new float[16];
        transform[0] = (float) result.mRotationMatrix[0];
        transform[4] = (float) result.mRotationMatrix[1];
        transform[8] = (float) result.mRotationMatrix[2];
        transform[12] = (float) result.mTranslation[0];

        transform[1] = (float) result.mRotationMatrix[3];
        transform[5] = (float) result.mRotationMatrix[4];
        transform[9] = (float) result.mRotationMatrix[5];
        transform[13] = (float) result.mTranslation[1];

        transform[2] = (float) result.mRotationMatrix[6];
        transform[6] = (float) result.mRotationMatrix[7];
        transform[10] = (float) result.mRotationMatrix[8];
        transform[14] = (float) result.mTranslation[2];

        transform[15] = 1.0f;

        mRenderableShader.use();

        GLES32.glBindVertexArray(mVAO);
        int transformLoc = GLES32.glGetUniformLocation(mRenderableShader.ID, "transform");

        GLES32.glUniformMatrix4fv(transformLoc, 1, false, FloatBuffer.wrap(transform));

        if (mType == RenderType.Axis)
            GLES32.glDrawElements(GLES32.GL_LINES, 6, GLES32.GL_UNSIGNED_INT, 0);
        else if(mType == RenderType.Point)
            GLES32.glDrawArrays(GLES32.GL_POINTS,0,mAmountOfIndices);
        else
            GLES32.glDrawElements(GLES32.GL_TRIANGLES, mAmountOfIndices, GLES32.GL_UNSIGNED_INT, 0);

        GLES32.glBindVertexArray(0);
    }

    /**
     * Vertex shader declaration
     */
    private final String vertexShaderSource = "#version 320 es\n"
            + "precision mediump float;\n"
            + "layout (location = 0) in vec3 aPos;\n"
            + "layout (location = 1) in vec4 aColor;\n"
            + "layout (location = 2) in vec2 aTexCoord;\n"
            + "out vec4 outColor;\n"
            + "out vec2 TexCoord;\n"
            + "uniform mat4 screenRotation;\n"
            + "uniform mat4 transform;\n"
            + "uniform mat4 perspective;\n"
            + "uniform mat4 scale;\n"
            + "uniform mat4 locPos;\n"
            + "void main()\n"
            + "{\n"
            + "   TexCoord = aTexCoord;\n"
            + "   outColor = aColor;\n"
            + "   gl_Position = screenRotation * perspective  * transform * locPos * scale * vec4(aPos, 1.0);\n"
            + "   gl_PointSize = 10.0;\n"
            + "}"; ///< Vertex shader code

    /**
     * Fragment shader declaration
     */
    private final String fragmentShaderSource = "#version 320 es\n"
            + "precision mediump float;\n"
            + "in vec4 outColor;\n"
            + "in vec2 TexCoord;\n"
            + "out vec4 glFragColor;\n"
            + "uniform sampler2D outTexture;\n"
            + "void main()\n"
            + "{\n"
            + "    if(TexCoord.x == 0.0 && TexCoord.y == 0.0)\n"
            + "        glFragColor = outColor;\n"
            + "    else\n"
            + "       glFragColor = texture(outTexture, TexCoord);"
            + "}\n"; ///< Fragment shader code

    RenderType mType = RenderType.Plane;  ///< Renderable type
    ShaderManager mRenderableShader;      ///< Shader instance which has shader codes
    int mVBO = 0;                         ///< Vertex Buffer Object ID
    int mVAO = 0;                         ///< Vertex Array Object ID
    int mEBO = 0;                         ///< Element Buffer Object ID
    int mTextureID = 0;                   ///< Texture ID
    int mAmountOfIndices = 0;             ///<

}
