package com.virnect.tracksample

import android.content.Context
import android.content.pm.ActivityInfo
import android.graphics.Typeface
import android.os.Bundle
import android.os.SystemClock
import android.util.Base64
import android.util.Log
import android.view.View
import android.widget.TextView
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.core.content.res.ResourcesCompat
import androidx.core.view.children
import com.virnect.Common
import com.virnect.FileAccessHelper
import com.virnect.Track
import kotlinx.android.synthetic.main.activity_track.*
import kotlin.experimental.xor

class TrackActivity : AppCompatActivity() {

    //region Initialization

    /**
     * Shared configuration singleton
     */
    private var trackConfig = TrackConfig.instance()

    /**
     * Log TAG
     */
    private val TAG = "TRACK_APP"

    /**
     * Reads the intent extra and sets the APP configuration accordingly
     */
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        this.setContentView(R.layout.activity_track)

        // Stop "map recording" button in bottom sheet
        buttonStopMapRecording.setOnClickListener {
            if (Track.saveMapTarget()) {
                mapRecorderLayout.visibility = View.GONE
                finish()
            }
        }
    }

    //endregion

    //region Android lifecycle integration

    /**
     * Initialize framework when the app starts
     */
    override fun onStart() {
        super.onStart()

        // CAD landscape mode
        if (trackConfig.playMode == PlayMode.CAD) {
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE)
        } else {
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT)
        }

        // Initialize track framework
        initializeTrackFramework(this)
    }

    /**
     * Start processing when the app is active
     */
    override fun onResume() {
        super.onResume()
        trackView.onResume()
        Track.startProcessing()

        // In case of fileSequence input, an external processing thread
        // is automatically started after initialization
        if (trackConfig.useFileSequence)
            startFileSequenceThread()
    }

    /**
     * Stop processing
     */
    override fun onPause() {
        super.onPause()
        trackView.onPause()
        Track.stopProcessing()
    }

    /**
     * Free all resources
     */
    override fun onStop() {
        super.onStop()
        Track.close()
    }

    //endregion

    //region Fullscreen and Backpress

    /**
     * Call fullscreen
     */
    override fun onWindowFocusChanged(hasFocus: Boolean) {
        super.onWindowFocusChanged(hasFocus)
        if (hasFocus) hideSystemUI()
    }

    /**
     * Apply fullscreen
     */
    @Suppress("DEPRECATION")
    private fun hideSystemUI() {
        window.decorView.systemUiVisibility = (
                View.SYSTEM_UI_FLAG_LAYOUT_STABLE
                        or View.SYSTEM_UI_FLAG_LAYOUT_HIDE_NAVIGATION
                        or View.SYSTEM_UI_FLAG_LAYOUT_FULLSCREEN
                        or View.SYSTEM_UI_FLAG_HIDE_NAVIGATION
                        or View.SYSTEM_UI_FLAG_FULLSCREEN
                        or View.SYSTEM_UI_FLAG_IMMERSIVE
                )
    }

    //endregion

    //region Track framework initialization

    /**
     * Initialize TRACK Framework
     * - Set License key
     * - Target setup:
     *      - Set Target names
     *      - Set Target data path
     *      - Optional - set file sequence path
     * - Initialize Framework
     * - Register processing callback
     * - Register visualization
     * - Start processing
     * @param context application main context
     */
    private fun initializeTrackFramework(context: Context?): Boolean {

        // Set license key
        if (!Track.setLicenseKey(
            )
        ) {
            Toast.makeText(
                context,
                "TRACK : ERROR No internet connection or license key is invalid",
                Toast.LENGTH_LONG
            ).show()
            return false
        }

        // Setup Target configuration for framework
        Track.setTargetDataPath(
            FileAccessHelper.getLocalPath(
                this,
                trackConfig.targetDataPath
            ).path
        )
        Track.setTargetNames(trackConfig.targets)

        // Playmode specific configuration
        // Hybrid, CAD & MapReplay do not need any special setup
        when (trackConfig.playMode) {
            PlayMode.MapRecording -> {
                trackView.isMapRecordMode = true
                Track.enableMapTargetGeneration(
                    true,
                    trackConfig.mapRecordingTargetName,
                    FileAccessHelper.getLocalPath(this, trackConfig.targetDataPathMap).path
                )
            }
            else -> {
            }
        }

        // In case of file sequence set path before initialization
        if (trackConfig.useFileSequence)
            Track.setFileSequencePath(trackConfig.fileSequencePath)

        // Initialize Track framework
        val success = Track.initialize(context)

        // Provide user feedback on failure
        if (!success) {
            Toast.makeText(
                context,
                "TRACK : ERROR could not initialize framework. " + Track.getFrameworkVersion(),
                Toast.LENGTH_LONG
            ).show()
            return false
        }

        // Set a callback function that is invoked after the process function
        getUICallbackFunction()?.also {
            Track.setProcessingCallback(it)
        }

        // Register targets for visualization
        if (trackConfig.playMode == PlayMode.MapRecording)
            trackView.setTargetNames(arrayOf(trackConfig.mapRecordingTargetName))
        else
            trackView.setTargetNames(trackConfig.targets)

        return success
    }

    /**
     * Starts a thread that calls the process function repeatedly.
     *
     * In case of file sequence usage the Track framework is not driven by the camera frames.
     * Therefore the process function of the framework needs to be called externally.
     */
    private fun startFileSequenceThread() {

        if (!trackConfig.useFileSequence) {
            Log.e(TAG, "Calling startFileSequenceThread without filesequence enabled ")
        }

        // Initialize track framework
        initializeTrackFramework(this)

        Thread {
            var runLoop = true
            while (runLoop) {
                try {
                    // loop processing with 60 FPS
                    runLoop = Track.processWithCallback()
                    SystemClock.sleep(16)
                } catch (exception: Exception) {
                    exception.message?.let { Log.e(TAG, it) }
                    runLoop = false
                }
            }
            val msg = "File sequence finished"
            Log.i(TAG, msg)
            Track.close()

            runOnUiThread {
                Toast.makeText(applicationContext, msg, Toast.LENGTH_LONG)
                    .show()
                finishAffinity()
            }
        }.start()
    }

    //endregion

    //region UI callback

    /**
     * Creates a TextView for each registered Target
     * Returns a callback that updates these UI elements depending on the tracking state
     */
    private fun getUICallbackFunction(): Runnable? {
        // Get standard target list
        var targetList = trackConfig.targets

        // Change in case of map recording
        if (trackConfig.playMode == PlayMode.MapRecording) {
            // Search for MAP target instead of registered ImageTarget (used for initialization)
            targetList = arrayOf(trackConfig.mapRecordingTargetName)

            // Show Instructions for map recording
            textViewMapRecordingInitializationTarget.text = trackConfig.targetMapRecording.get(0)
            bottomSheetStartMapRecording.visibility = View.VISIBLE
            bottomSheetStopMapRecording.visibility = View.INVISIBLE

            // Show map recording UI container
            mapRecorderLayout.visibility = View.VISIBLE
        } else
            mapRecorderLayout.visibility = View.GONE

        // Create UI and callback function for targets
        targetList?.also { targetsArray ->
            // Prepare target result array
            val trackerResults = targetsArray.map { Common.TrackerResult(it) }.toTypedArray()

            // Prepare corresponding UI elements
            targetsArray.map {
                if (getTargetUI(it) == null) {
                    val targetText = TextView(applicationContext)
                    targetText.tag = it
                    targetText.typeface = ResourcesCompat.getFont(this, R.font.roboto_mono)
                    targetText.setTypeface(targetText.typeface, Typeface.NORMAL)
                    targetText.textSize = 18f
                    targetContainer.addView(targetText)
                }
            }

            // This function will be called after each process call of the TRACK framework
            return Runnable {
                // Retrieve results
                Track.getTrackingResult(trackerResults)

                // Display tracking result as text
                runOnUiThread {
                    for (i in trackerResults.indices) {
                        val result = trackerResults[i]
                        getTargetUI(result.mTargetName)?.also {
                            it.text = "%s - %s ".format(
                                result.mTargetName.padEnd(12),
                                if (result.mStatus == Common.TrackingState.TRACKED) "TRACKED" else "SEARCHING"
                            )
                            it.setTextColor(
                                ResourcesCompat.getColor(
                                    resources,
                                    if (result.mStatus == Common.TrackingState.TRACKED) R.color.colorGreenText else R.color.colorRedText,
                                    null
                                )
                            )

                            // Map recording mode behaviour
                            if (trackConfig.playMode == PlayMode.MapRecording && result.mStatus == Common.TrackingState.TRACKED) {
                                // If MAP is tracked once, hide map recording instructions, show save button
                                bottomSheetStartMapRecording.visibility = View.INVISIBLE
                                bottomSheetStopMapRecording.visibility = View.VISIBLE

                                // Lock rotation
                                setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LOCKED)
                            }
                        }
                    }
                }

                // Update visualization
                trackView.requestRender()
            }
        }

        // Return null if there are no targets
        Log.e(TAG, "Could not create UI callback - no targets")
        return null
    }

    /**
     * Helper function to get Target UI by tag
     */
    private fun getTargetUI(tag: String): TextView? {
        for (elem in targetContainer.children) {
            if (elem.tag != null && elem.tag.equals(tag))
                return elem as TextView
        }
        return null
    }

    //endregion
}