package com.virnect.tracksample;

import android.content.Context;
import android.opengl.EGL14;
import android.opengl.GLSurfaceView;
import android.util.AttributeSet;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.Surface;
import android.view.SurfaceHolder;

import androidx.annotation.NonNull;

import com.virnect.Common;
import com.virnect.Track;
import com.virnect.tracksample.Renderer.Renderable;

import java.util.HashMap;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

/**
 * Example viewer class
 * Basic functions are implemented
 * You need to implement AutoFitViewer function and Renderer using GL or other 3rdParties
 */
public class TrackRenderView extends GLSurfaceView implements GLSurfaceView.Renderer {

    private final String TAG = this.getClass().getName(); ///< Log tag

    private final HashMap<String, SampleTarget> renderableLists = new HashMap<>();          ///< Target lists with Renderable
    private final Renderable backgroundRenderer = new Renderable();                         ///< Background object
    private Common.CameraCalibration cameraCalibration = new Common.CameraCalibration();    ///< Device camera calibration (gets set with renderables)
    private String[] targetNames = new String[0];                                           ///< buffer to save target names in case of configuration change
    private Common.TrackerResult[] trackerResults = new Common.TrackerResult[0];            ///< TrackerResult array for fetching tracking result
    private Common.TargetInformation[] targetInformation = new Common.TargetInformation[0]; ///< TrackerInformation array for fetching tracker information
    public boolean isMapRecordMode = false;                                                 ///< Flag for checking record mode

    /**
     * TrackRenderView constructor
     * This constructor has GLSurfaceView constructor parameter
     *
     * @param context Application's main context
     * @param attrs   View's attributes
     */
    public TrackRenderView(Context context, AttributeSet attrs) {
        super(context, attrs);

        setEGLContextClientVersion(3);
        setRenderer(this);
        setRenderMode(RENDERMODE_WHEN_DIRTY);

        getHolder().addCallback(this);

        //Apply screen rotation according to display rotation
        float angle = 0;
        switch (context.getDisplay().getRotation()) {
            case Surface.ROTATION_0:
                angle = -90;
                break;
            case Surface.ROTATION_90:
                angle = 0;
                break;
            case Surface.ROTATION_180:
                angle = 90;
                break;
            case Surface.ROTATION_270:
                angle = 180;
                break;
        }
        Renderable.applyScreenRotation(angle);

        Log.d(TAG, "CONTEXT: " + context);
    }

    /**
     * Set target list to visualize
     *
     * @param targetNames Tracking target names
     */
    public void setTargetNames(String[] targetNames) {
        this.targetNames = targetNames;
    }

    /**
     * Create SampleTargets
     */
    private void createTargets() {
        trackerResults = new Common.TrackerResult[targetNames.length];
        targetInformation = new Common.TargetInformation[targetNames.length];

        for (int i = 0; i < targetNames.length; i++) {
            trackerResults[i] = new Common.TrackerResult(targetNames[i]);
            targetInformation[i] = new Common.TargetInformation(targetNames[i]);
        }

        Track.getCameraCalibration(cameraCalibration);
        Track.getTargetInfo(targetInformation);

        // Initialize targets on GL thread
        queueEvent(new Runnable() {
            @Override
            public void run() {
                if (targetInformation != null)
                    renderableLists.clear();

                for (Common.TargetInformation info : targetInformation) {
                    // Assume that record mode has only one target
                    if (info.getType() == Common.TargetType.MAP.ordinal() || isMapRecordMode) {
                        SampleTarget target = new SampleTarget(Renderable.RenderType.Point, cameraCalibration, new double[]{1, 1, 1});
                        target.setVertices(Track.getMapPoints(info.getName()));
                        renderableLists.put(info.getName(), target);
                    } else if (info.getType() == Common.TargetType.CADMODEL.ordinal()) {
                        SampleTarget target = new SampleTarget(Renderable.RenderType.Model, cameraCalibration, new double[]{1, 1, 1});
                        Common.CADModel model = new Common.CADModel();
                        Track.getCADModel(model);
                        target.setModel(model);
                        renderableLists.put(info.getName(), target);
                    } else
                        renderableLists.put(info.getName(), new SampleTarget(Renderable.RenderType.Plane, cameraCalibration, info.getMax()));
                }
            }
        });
    }

    /**
     * Resize view for auto fit
     *
     * @param widthMeasureSpec  Display width
     * @param heightMeasureSpec Display height
     */
    @Override
    protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
        super.onMeasure(widthMeasureSpec, heightMeasureSpec);

        int width = MeasureSpec.getSize(widthMeasureSpec);
        int height = MeasureSpec.getSize(heightMeasureSpec);

        int[] previewSize = Track.getPreviewSize();

        if (previewSize[0] == 0) return;

        float previewRatio = (float) previewSize[0] / previewSize[1];
        float screenRatio = (float) width / height;

        if (previewRatio < screenRatio) {
            float measureRatio = previewRatio / screenRatio;
            float newHeight = height / measureRatio;
            setMeasuredDimension(width, (int) newHeight);
            setY((newHeight - height) * -0.5f);
        }
    }

    /**
     * Moves the view to a negative X coordinate to center the view on the screen
     *
     * @param changed This is a new size or position for this view
     * @param left    Left position, relative to parent
     * @param top     Top position, relative to parent
     * @param right   Right position, relative to parent
     * @param bottom  Bottom position, relative to parent
     */
    @Override
    protected void onLayout(boolean changed, int left, int top, int right, int bottom) {
        super.onLayout(changed, left, top, right, bottom);
        if (!changed || left != 0 || top != 0)
            return;

        boolean landscape = false;
        switch (getContext().getDisplay().getRotation()) {
            case Surface.ROTATION_0:
                break;
            case Surface.ROTATION_90:
                landscape = true;
                break;
            case Surface.ROTATION_180:
                break;
            case Surface.ROTATION_270:
                landscape = true;
                break;
        }

        // Center view on screen (constraint layout does not allow negative initial poses)
        DisplayMetrics screenSize = getContext().getResources().getDisplayMetrics();
        if (landscape) {
            int offsetY = (bottom - screenSize.heightPixels) / 2;
            setTranslationY(-offsetY);
        } else {
            int offsetX = (right - screenSize.widthPixels) / 2;
            setTranslationX(-offsetX);
        }
    }

    /**
     * This is called immediately after the surface is first created.
     * Implementations of this should start up whatever rendering code
     * they desire.  Note that only one thread can ever draw into
     * a {@link Surface}, so you should not draw into the Surface here
     * if your normal rendering will be in another thread.
     *
     * @param holder The SurfaceHolder whose surface is being created.
     */
    @Override
    public void surfaceCreated(@NonNull SurfaceHolder holder) {
        super.surfaceCreated(holder);
    }

    /**
     * This is called immediately before a surface is being destroyed. After
     * returning from this call, you should no longer try to access this
     * surface.  If you have a rendering thread that directly accesses
     * the surface, you must ensure that thread is no longer touching the
     * Surface before returning from this function.
     *
     * @param holder The SurfaceHolder whose surface is being destroyed.
     */
    @Override
    public void surfaceDestroyed(@NonNull SurfaceHolder holder) {
        super.surfaceDestroyed(holder);
    }

    /**
     * Called when the surface is created or recreated.
     * <p>
     * Called when the rendering thread
     * starts and whenever the EGL context is lost. The EGL context will typically
     * be lost when the Android device awakes after going to sleep.
     * <p>
     * Since this method is called at the beginning of rendering, as well as
     * every time the EGL context is lost, this method is a convenient place to put
     * code to create resources that need to be created when the rendering
     * starts, and that need to be recreated when the EGL context is lost.
     * Textures are an example of a resource that you might want to create
     * here.
     * <p>
     * Note that when the EGL context is lost, all OpenGL resources associated
     * with that context will be automatically deleted. You do not need to call
     * the corresponding "glDelete" methods such as glDeleteTextures to
     * manually delete these lost resources.
     * <p>
     *
     * @param gl     the GL interface. Use <code>instanceof</code> to
     *               test if the interface supports GL11 or higher interfaces.
     * @param config the EGLConfig of the created surface. Can be used
     */
    @Override
    public void onSurfaceCreated(GL10 gl, EGLConfig config) {
        gl.glEnable(gl.GL_BLEND);
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA);

        Track.getCameraCalibration(cameraCalibration);

        backgroundRenderer.initialize(Renderable.RenderType.Plane, true);
        backgroundRenderer.setOrthogonal();
        backgroundRenderer.setScale(1.0f, 1.0f, 1.0f);
        backgroundRenderer.setPosition(0, 0, 0);

        // Create SampleTargets for visualizing
        createTargets();
    }

    /**
     * Called when the surface changed size.
     * <p>
     * Called after the surface is created and whenever
     * the OpenGL ES surface size changes.
     * <p>
     * Typically you will set your viewport here. If your camera
     * is fixed then you could also set your projection matrix here:
     * <pre class="prettyprint">
     * void onSurfaceChanged(GL10 gl, int width, int height) {
     *     gl.glViewport(0, 0, width, height);
     *     // for a fixed camera, set the projection too
     *     float ratio = (float) width / height;
     *     gl.glMatrixMode(GL10.GL_PROJECTION);
     *     gl.glLoadIdentity();
     *     gl.glFrustumf(-ratio, ratio, -1, 1, 1, 10);
     * }
     * </pre>
     *
     * @param gl     the GL interface. Use <code>instanceof</code> to
     *               test if the interface supports GL11 or higher interfaces.
     * @param width  the GL surface width
     * @param height the GL surface height
     */
    @Override
    public void onSurfaceChanged(GL10 gl, int width, int height) {
        gl.glViewport(0, 0, width, height);
    }

    /**
     * Called to draw the current frame.
     * <p>
     * This method is responsible for drawing the current frame.
     * <p>
     * The implementation of this method typically looks like this:
     * <pre class="prettyprint">
     * void onDrawFrame(GL10 gl) {
     *     gl.glClear(GL10.GL_COLOR_BUFFER_BIT | GL10.GL_DEPTH_BUFFER_BIT);
     *     //... other gl calls to render the scene ...
     * }
     * </pre>
     *
     * @param gl the GL interface. Use <code>instanceof</code> to
     *           test if the interface supports GL11 or higher interfaces.
     */
    @Override
    public void onDrawFrame(GL10 gl) {
        EGL14.eglGetCurrentContext();
        backgroundRenderer.render(Track.getPreviewSize(), Track.getFrame());

        Track.getTrackingResult(trackerResults);
        if (trackerResults != null)
            for (Common.TrackerResult result : trackerResults)
                if (renderableLists.containsKey(result.mTargetName) && result.mStatus == Common.TrackingState.TRACKED) {
                    SampleTarget sampleTarget = renderableLists.get(result.mTargetName);

                    if (sampleTarget == null) return;

                    if (isMapRecordMode && sampleTarget.updateVersion(Track.getMapPointsVersion(result.mTargetName)))
                        sampleTarget.setVertices(Track.getMapPoints(result.mTargetName));

                    sampleTarget.render(result);
                }
    }

    /**
     * Target renderable class for rendering plane and axis
     */
    static class SampleTarget {
        private final Renderable targetRenderer = new Renderable();
        private final Renderable axisRenderer = new Renderable();
        private int version = 0;

        /**
         * SamplePlaneTarget's constructor
         *
         * @param cameraCalibration CameraCalibration information for setting projection matrix
         * @param size              Target scale
         */
        SampleTarget(final Renderable.RenderType type, final Common.CameraCalibration cameraCalibration, double[] size) {
            targetRenderer.initialize(type, false);
            targetRenderer.setPerspective(cameraCalibration);
            targetRenderer.setScale((float) size[0], (float) size[1], (float) size[2]);
            targetRenderer.setPosition(0, 0, 0);

            axisRenderer.initialize(Renderable.RenderType.Axis, false);
            axisRenderer.setPerspective(cameraCalibration);

            if (type == Renderable.RenderType.Model)
                axisRenderer.setScale(0f, 0f, 0f);
            else if (type == Renderable.RenderType.Point)
                axisRenderer.setScale(0.1f, 0.1f, 0.1f);
            else
                axisRenderer.setScale((float) size[1], (float) size[1], (float) size[1]);

            axisRenderer.setPosition(0, 0, 0);
        }

        /**
         * Check sample target version and update if needed
         *
         * @param newVersion New version for checing
         * @return If current version is updated or not
         */
        public boolean updateVersion(int newVersion) {
            if (version < newVersion) {
                version = newVersion;
                return true;
            }

            return false;
        }

        /**
         * Set vertices of SampleTarget
         *
         * @param vertices vertices for drawing
         */
        public void setVertices(Common.Vec3[] vertices) {
            targetRenderer.setVertices(vertices);
        }

        public void setModel(Common.CADModel model) {
            targetRenderer.setVertices(model.mVertices, false);
            targetRenderer.setIndices(model.mIndices);
        }

        /**
         * Render this target
         *
         * @param result Tracking result struct of this target
         */
        public void render(final Common.TrackerResult result) {
            targetRenderer.render(result);
            axisRenderer.render(result);
        }
    }
}